# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the qa workflow."""

from collections.abc import Sequence
from typing import Any, ClassVar, override

from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestResultStatus,
    DebianAutopkgtestSource,
    DebianBinaryPackage,
    DebianLintian,
    DebianLintianSeverity,
    DebianLintianSummary,
    DebianSourcePackage,
    TaskTypes,
    get_source_package_name,
    get_source_package_version,
)
from debusine.db.models import Artifact, WorkRequest
from debusine.server.collections import DebianSuiteManager
from debusine.server.workflows import (
    QAWorkflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.models import (
    BaseWorkflowData,
    QAWorkflowData,
    QAWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    BackendType,
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LintianFailOnSeverity,
    LookupMultiple,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class QAWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestCase[QAWorkflow]
):
    """Unit tests for :py:class:`QAWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "bookworm",
        "binary_artifacts": ["internal@collections/name:build-amd64"],
    }

    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]
    package_build_log: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact(
                architecture="amd64"
            )
        )
        cls.package_build_log = cls.playground.create_build_log_artifact()
        cls.playground.create_artifact_relation(
            artifact=cls.package_build_log, target=cls.binary_artifact
        )

        # Create artifacts needed to satisfy task data lookups
        cls.playground.create_debian_environment(
            codename="sid", architecture="amd64", variant="sbuild"
        )
        cls.playground.create_debian_environment(
            codename="bookworm", architecture="amd64", variant="autopkgtest"
        )
        cls.playground.create_debian_environment(
            codename="bookworm", architecture="amd64", variant="lintian"
        )
        cls.playground.create_debian_environment(
            codename="bookworm", architecture="amd64", variant="blhc"
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.pk
        self.internal_collection_artifacts = [
            (self.binary_artifact, {"name": "build-amd64"}),
            (self.package_build_log, {"name": "buildlog-amd64"}),
        ]

    def orchestrate(
        self,
        extra_data: dict[str, Any],
        *,
        architectures: Sequence[str] | None = None,
        sbuild_workflow_source_artifact: Artifact,
        internal_collection_artifacts: (
            list[tuple[Artifact, dict[str, Any]]] | None
        ) = None,
    ) -> WorkRequest:
        """Create a QAWorkflow and call orchestrate_workflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Example workflow."""

            @override
            def populate(self_) -> None:
                source_artifact_id = (
                    sbuild_workflow_source_artifact.id
                    if sbuild_workflow_source_artifact is not None
                    else self.source_artifact.id
                )
                sbuild = self_.work_request.create_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact_id),
                        target_distribution="debian:sid",
                        architectures=list(architectures or ["all"]),
                    ),
                )
                self.playground.advance_work_request(sbuild, mark_running=True)

                if architectures is not None:
                    for arch in architectures:
                        child = sbuild.create_child_worker(
                            task_name="sbuild",
                            task_data=SbuildData(
                                input=SbuildInput(
                                    source_artifact=source_artifact_id
                                ),
                                build_architecture=arch,
                                environment="debian/match:codename=sid",
                            ),
                            workflow_data=WorkRequestWorkflowData(
                                display_name=f"Build {arch}",
                                step=f"build-{arch}",
                            ),
                        )

                        self_.provides_artifact(
                            child,
                            ArtifactCategory.UPLOAD,
                            f"build-{arch}",
                            data={
                                "binary_names": ["hello"],
                                "architecture": arch,
                                "source_package_name": "hello",
                            },
                        )
                        self_.provides_artifact(
                            child,
                            ArtifactCategory.PACKAGE_BUILD_LOG,
                            f"buildlog-{arch}",
                            data={
                                "architecture": arch,
                                "source_package_name": "hello",
                                "source_package_version": "1.0-1",
                            },
                        )

                sbuild.unblock_workflow_children()

                data = {
                    "source_artifact": source_artifact_id,
                    "binary_artifacts": [self.binary_artifact.id],
                    "enable_blhc": False,
                    "vendor": "debian",
                    "codename": "bookworm",
                }

                qa, _ = self_.work_request_ensure_child_workflow(
                    task_name="qa",
                    task_data=QAWorkflowData(**{**data, **extra_data}),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="DebDiff workflow", step="debdiff-workflow"
                    ),
                )
                self_.orchestrate_child(qa)

        root = self.playground.create_workflow(
            task_name="examplepipeline",
            internal_collection_artifacts=internal_collection_artifacts,
        )
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root.children.get(task_name="qa")

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        self.task.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        self.validate = False
        self.task_data["qa_suite"] = "nonexistent@debian:suite"
        self.task_data["enable_reverse_dependencies_autopkgtest"] = True
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            QAWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact.id],
            ),
        )

    def test_compute_dynamic_data_reference_source_artifact(self) -> None:
        reference_source_artifact = self.source_artifact
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )

        self.task_data["source_artifact"] = new_source_artifact.id
        self.task_data["reference_source_artifact"] = (
            reference_source_artifact.id
        )
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            QAWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-2",
                source_artifact_id=new_source_artifact.id,
                reference_source_artifact_id=reference_source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact.id],
            ),
        )

    def test_compute_dynamic_data_source_artifact_wrong_category(self) -> None:
        """`source_artifact` must be a source-package or upload."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        self.task_data["source_artifact"] = artifact.id
        with self.assertRaisesRegex(
            TaskConfigError,
            r"^source_artifact: unexpected artifact category: 'debusine:test'. "
            r"Valid categories: \['debian:source-package', 'debian:upload'\]$",
        ):
            self.task

    def test_get_input_artifacts_mandatory_data(self) -> None:
        upload_artifacts = self.playground.create_upload_artifacts()
        self.task_data["source_artifact"] = upload_artifacts.upload.id
        self.task_data["enable_blhc"] = False
        self.assertIsNone(self.task.dynamic_data)
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=None,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=None,
                ),
            ],
        )

        self.task.dynamic_data = self.compute_dynamic_data(self.task)
        self.assertIsNotNone(self.task.dynamic_data)

        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=upload_artifacts.source.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[self.binary_artifact.id],
                ),
            ],
        )

    def test_get_input_artifacts_with_optional_data(self) -> None:
        upload_artifacts = self.playground.create_upload_artifacts()
        reference_source_artifact = self.source_artifact

        self.task_data["source_artifact"] = upload_artifacts.upload.id
        self.task_data["reference_source_artifact"] = (
            reference_source_artifact.id
        )
        self.task_data["enable_blhc"] = False
        self.task.dynamic_data = QAWorkflowDynamicData(
            source_artifact_id=upload_artifacts.source.id,
            binary_artifacts_ids=[self.binary_artifact.id],
            reference_source_artifact_id=reference_source_artifact.id,
        )

        assert self.task.data.reference_source_artifact is not None
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=upload_artifacts.source.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=self.task.dynamic_data.binary_artifacts_ids,
                ),
                InputArtifactSingle(
                    lookup=self.task.data.reference_source_artifact,
                    label="reference_source_artifact",
                    artifact_id=reference_source_artifact.id,
                ),
            ],
        )

    def test_source_package_upload(self) -> None:
        """`source_artifact` can be an upload."""
        upload_artifacts = self.playground.create_upload_artifacts()
        self.task_data["source_artifact"] = upload_artifacts.upload.id
        self.task_data["enable_blhc"] = False
        self.assertEqual(
            workflow_utils.source_package(self.task), upload_artifacts.source
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """Test populate."""
        architectures = ["amd64"]
        source_artifact = self.source_artifact

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "architectures": ["amd64"],
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            },
            architectures=architectures,
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )

        assert workflow.parent is not None
        self.assertEqual(autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            autopkgtest.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            },
        )
        self.assertEqual(
            autopkgtest.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "autopkgtest",
                "step": "autopkgtest",
            },
        )
        self.assertQuerySetEqual(autopkgtest.dependencies.all(), [])

        # AutopkgtestWorkflow.populate() was called and created its tasks
        self.assertQuerySetEqual(
            autopkgtest.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "autopkgtest",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

        self.assertFalse(
            workflow.children.filter(
                task_name="reverse_dependencies_autopkgtest",
                task_type=TaskTypes.WORKFLOW,
            ).exists()
        )

        lintian = workflow.children.get(
            task_name="lintian", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(lintian.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            lintian.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "fail_on_severity": LintianFailOnSeverity.ERROR,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )
        self.assertEqual(
            lintian.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "lintian",
                "step": "lintian",
            },
        )
        self.assertQuerySetEqual(lintian.dependencies.all(), [])
        # LintianWorkflow.populate() was called and created its tasks
        self.assertQuerySetEqual(
            lintian.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "lintian",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

        piuparts = workflow.children.get(
            task_name="piuparts", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(piuparts.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            piuparts.task_data,
            {
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            },
        )

        self.assertEqual(
            piuparts.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "piuparts",
                "step": "piuparts",
            },
        )
        self.assertQuerySetEqual(piuparts.dependencies.all(), [])

        # PiupartsWorkflow.populate() was called and created its tasks
        self.assertQuerySetEqual(
            piuparts.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "piuparts",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

    @preserve_db_task_registry()
    def test_populate_use_available_architectures(self) -> None:
        """
        Test populate uses available architectures.

        The user didn't specify "architectures", so QAWorkflow checks
        available architectures and "all" and uses them.
        """
        for arch in ["amd64", "i386"]:
            self.playground.create_debian_environment(architecture=arch)
        source_artifact = self.source_artifact
        binary_artifact = self.binary_artifact

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [binary_artifact.id],
            },
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(
            autopkgtest.task_data["architectures"], ["all", "amd64", "i386"]
        )

    @preserve_db_task_registry()
    def test_populate_disable_autopkgtest_lintian_piuparts(self) -> None:
        """Populate does not create autopkgtest, lintian nor piuparts."""
        source_artifact = self.source_artifact
        binary_artifact = self.binary_artifact
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [binary_artifact.id],
                "architectures": ["amd64"],
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": False,
            },
            sbuild_workflow_source_artifact=source_artifact,
        )

        self.assertFalse(
            workflow.children.filter(
                task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
            ).exists()
        )
        self.assertFalse(
            workflow.children.filter(
                task_name="reverse_dependencies_autopkgtest",
                task_type=TaskTypes.WORKFLOW,
            ).exists()
        )
        self.assertFalse(
            workflow.children.filter(
                task_name="lintian", task_type=TaskTypes.WORKFLOW
            ).exists()
        )
        self.assertFalse(
            workflow.children.filter(
                task_name="piuparts", task_type=TaskTypes.WORKFLOW
            ).exists()
        )

    def test_populate_uploads(self) -> None:
        """The workflow accepts debian:upload artifacts in all inputs."""
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        architectures = ["amd64", "i386"]

        wr = self.playground.create_workflow(
            task_name="qa",
            task_data=QAWorkflowData(
                source_artifact=f"{artifacts.upload.id}@artifacts",
                binary_artifacts=LookupMultiple((artifacts.upload.id,)),
                architectures=architectures,
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                enable_autopkgtest=True,
                enable_reverse_dependencies_autopkgtest=True,
                enable_debdiff=True,
                enable_lintian=True,
                enable_piuparts=True,
                enable_blhc=True,
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.order_by("id").all())
        expected_src_artifact = f"{artifacts.upload.id}@artifacts"
        expected_bin_artifacts = [artifacts.upload.id]
        expected: dict[str, dict[str, Any]] = {
            "autopkgtest": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
            "blhc": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
            "debdiff": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
            "lintian": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
            "piuparts": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
            "reverse_dependencies_autopkgtest": {
                "source_artifact": expected_src_artifact,
                "binary_artifacts": expected_bin_artifacts,
            },
        }
        self.assertEqual(len(children), len(expected))
        for child in children:
            with self.subTest(task_name=child.task_name):
                expected_data = expected[child.task_name]
                for k, v in expected_data.items():
                    self.assertEqual(
                        child.task_data.get(k), v, f"task_data[{k}]"
                    )

    @preserve_db_task_registry()
    def test_populate_enable_reverse_dependencies_autopkgtest(self) -> None:
        """Populate can enable reverse_dependencies_autopkgtest."""
        architectures = ["amd64"]
        source_artifact = self.playground.create_source_artifact(name="hello")
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        dep_source_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="depends",
                version="1",
                type="dpkg",
                dsc_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Testsuite": "autopkgtest",
                },
            ),
        )
        sid.manager.add_artifact(
            dep_source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        dep_binary_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="depends",
                srcpkg_version="1",
                deb_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Architecture": "all",
                    "Depends": "hello",
                },
                deb_control_files=[],
            ),
        )
        sid.manager.add_artifact(
            dep_binary_artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "architectures": ["amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "enable_reverse_dependencies_autopkgtest": True,
            },
            architectures=architectures,
            sbuild_workflow_source_artifact=source_artifact,
        )

        rdep_autopkgtest = workflow.children.get(
            task_name="reverse_dependencies_autopkgtest",
            task_type=TaskTypes.WORKFLOW,
        )

        assert workflow.parent is not None
        self.assertEqual(rdep_autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            rdep_autopkgtest.task_data,
            {
                "prefix": "",
                "reference_prefix": "",
                "source_artifact": source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": None,
                "enable_regression_tracking": False,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "codename": "bookworm",
                "backend": BackendType.AUTO,
                "architectures": ["amd64"],
                "arch_all_build_architecture": "amd64",
                "extra_repositories": None,
            },
        )
        self.assertEqual(
            rdep_autopkgtest.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "autopkgtests of reverse-dependencies",
                "step": "reverse-dependencies-autopkgtest",
            },
        )
        self.assertQuerySetEqual(rdep_autopkgtest.dependencies.all(), [])

        # ReverseDependenciesAutopkgtestWorkflow.populate() was called and
        # created its sub-workflow.
        autopkgtest = rdep_autopkgtest.children.get()
        self.assertEqual(autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            autopkgtest.task_data["source_artifact"],
            "sid@debian:suite/source-version:depends_1",
        )
        self.assertQuerySetEqual(
            autopkgtest.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "autopkgtest",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

    @preserve_db_task_registry()
    def test_populate_update_qa_results(self) -> None:
        """Populate can update QA results."""
        architectures = ["amd64"]
        source_artifact = self.source_artifact
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        self.playground.create_collection(
            name="sid", category=CollectionCategory.QA_RESULTS
        )
        dep_source_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="depends",
                version="1",
                type="dpkg",
                dsc_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Testsuite": "autopkgtest",
                },
            ),
        )
        sid.manager.add_artifact(
            dep_source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        dep_binary_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="depends",
                srcpkg_version="1",
                deb_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Architecture": "all",
                    "Depends": "hello",
                },
                deb_control_files=[],
            ),
        )
        sid.manager.add_artifact(
            dep_binary_artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        workflow = self.orchestrate(
            extra_data={
                "prefix": "reference-qa-result|",
                "reference_prefix": "",
                "source_artifact": source_artifact.id,
                # This would normally be the corresponding binary artifacts
                # in qa_suite, but that doesn't really matter for testing
                # purposes.
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "architectures": ["amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": False,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "enable_reverse_dependencies_autopkgtest": True,
                "enable_blhc": True,
            },
            architectures=architectures,
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )

        assert workflow.parent is not None
        self.assertEqual(autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            autopkgtest.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "reference-qa-result|",
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_prefix": "",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
                "extra_repositories": None,
            },
        )
        self.assertEqual(
            autopkgtest.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "autopkgtest",
                "step": "autopkgtest",
            },
        )
        self.assertQuerySetEqual(autopkgtest.dependencies.all(), [])

        # AutopkgtestWorkflow.populate() was called and created its tasks.
        self.assertQuerySetEqual(
            autopkgtest.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "autopkgtest",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

        rdep_autopkgtest = workflow.children.get(
            task_name="reverse_dependencies_autopkgtest",
            task_type=TaskTypes.WORKFLOW,
        )

        assert workflow.parent is not None
        self.assertEqual(rdep_autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            rdep_autopkgtest.task_data,
            {
                "prefix": "reference-qa-result|",
                "reference_prefix": "",
                "source_artifact": source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": False,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
                "codename": "bookworm",
                "backend": BackendType.AUTO,
                "architectures": ["amd64"],
                "arch_all_build_architecture": "amd64",
                "extra_repositories": None,
            },
        )
        self.assertEqual(
            rdep_autopkgtest.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "autopkgtests of reverse-dependencies",
                "step": "reverse-dependencies-autopkgtest",
            },
        )
        self.assertQuerySetEqual(rdep_autopkgtest.dependencies.all(), [])

        # ReverseDependenciesAutopkgtestWorkflow.populate() was called and
        # created its sub-workflow.
        sub_autopkgtest = rdep_autopkgtest.children.get()
        self.assertEqual(sub_autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            sub_autopkgtest.task_data["source_artifact"],
            "sid@debian:suite/source-version:depends_1",
        )
        self.assertQuerySetEqual(
            sub_autopkgtest.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "autopkgtest",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

        lintian = workflow.children.get(
            task_name="lintian", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(lintian.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            lintian.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "fail_on_severity": LintianFailOnSeverity.ERROR,
                "prefix": "reference-qa-result|",
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_prefix": "",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
            },
        )
        self.assertEqual(
            lintian.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "lintian",
                "step": "lintian",
            },
        )
        self.assertQuerySetEqual(lintian.dependencies.all(), [])

        # LintianWorkflow.populate() was called and created its tasks.
        self.assertQuerySetEqual(
            lintian.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "lintian",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                )
            ],
        )

        piuparts = workflow.children.get(
            task_name="piuparts", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(piuparts.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            piuparts.task_data,
            {
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "reference-qa-result|",
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_prefix": "",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
                "extra_repositories": None,
            },
        )

        self.assertEqual(
            piuparts.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "piuparts",
                "step": "piuparts",
            },
        )
        self.assertQuerySetEqual(piuparts.dependencies.all(), [])

        blhc = workflow.children.get(
            task_name="blhc", task_type=TaskTypes.WORKFLOW
        )

        assert workflow.parent is not None
        self.assertEqual(blhc.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(
            blhc.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "reference-qa-result|",
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_prefix": "",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
            },
        )
        self.assertEqual(
            blhc.workflow_data_json,
            {
                "allow_failure": False,
                "display_name": "build log hardening check",
                "step": "blhc",
            },
        )
        self.assertQuerySetEqual(
            blhc.dependencies.values_list("task_type", "task_name"),
            [(TaskTypes.WORKER, "sbuild")],
        )
        # No children yet; see test_populate_blhc.
        self.assertFalse(blhc.children.exists())

    @preserve_db_task_registry()
    def test_populate_enable_regression_tracking(self) -> None:
        """Populate can run in a mode that tracks regressions."""
        architectures = ["amd64"]
        source_artifact = self.source_artifact
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        self.playground.create_collection(
            name="sid", category=CollectionCategory.QA_RESULTS
        )
        dep_source_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="depends",
                version="1",
                type="dpkg",
                dsc_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Testsuite": "autopkgtest",
                },
            ),
        )
        sid.manager.add_artifact(
            dep_source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        dep_binary_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="depends",
                srcpkg_version="1",
                deb_fields={
                    "Package": "depends",
                    "Version": "1",
                    "Architecture": "all",
                    "Depends": "hello",
                },
                deb_control_files=[],
            ),
        )
        sid.manager.add_artifact(
            dep_binary_artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        workflow = self.orchestrate(
            extra_data={
                "reference_prefix": "reference-qa-result|",
                "source_artifact": source_artifact.id,
                # This would normally be the corresponding binary artifacts
                # in qa_suite, but that doesn't really matter for testing
                # purposes.
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "architectures": ["amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": True,
                "enable_reverse_dependencies_autopkgtest": True,
                "enable_blhc": True,
            },
            architectures=architectures,
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertDictContainsAll(
            autopkgtest.task_data,
            {
                "enable_regression_tracking": True,
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            autopkgtest.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "autopkgtest",
                "step": "autopkgtest",
            },
        )
        self.assertQuerySetEqual(autopkgtest.dependencies.all(), [])
        self.assertTrue(autopkgtest.children.exists())

        rdep_autopkgtest = workflow.children.get(
            task_name="reverse_dependencies_autopkgtest",
            task_type=TaskTypes.WORKFLOW,
        )
        self.assertEqual(rdep_autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertDictContainsAll(
            rdep_autopkgtest.task_data,
            {
                "enable_regression_tracking": True,
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            rdep_autopkgtest.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "autopkgtests of reverse-dependencies",
                "step": "reverse-dependencies-autopkgtest",
            },
        )
        self.assertQuerySetEqual(rdep_autopkgtest.dependencies.all(), [])

        # ReverseDependenciesAutopkgtestWorkflow.populate() was called and
        # created its sub-workflow.
        sub_autopkgtest = rdep_autopkgtest.children.get()
        self.assertEqual(sub_autopkgtest.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            sub_autopkgtest.task_data["source_artifact"],
            "sid@debian:suite/source-version:depends_1",
        )
        self.assertTrue(sub_autopkgtest.children.exists())

        lintian = workflow.children.get(
            task_name="lintian", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(lintian.status, WorkRequest.Statuses.RUNNING)
        self.assertDictContainsAll(
            lintian.task_data,
            {
                "enable_regression_tracking": True,
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            lintian.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "lintian",
                "step": "lintian",
            },
        )
        self.assertQuerySetEqual(lintian.dependencies.all(), [])
        self.assertTrue(lintian.children.exists())

        piuparts = workflow.children.get(
            task_name="piuparts", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(piuparts.status, WorkRequest.Statuses.RUNNING)
        self.assertDictContainsAll(
            piuparts.task_data,
            {
                "enable_regression_tracking": True,
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            piuparts.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "piuparts",
                "step": "piuparts",
            },
        )
        self.assertQuerySetEqual(piuparts.dependencies.all(), [])
        self.assertTrue(piuparts.children.exists())

        blhc = workflow.children.get(
            task_name="blhc", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(blhc.status, WorkRequest.Statuses.BLOCKED)
        self.assertDictContainsAll(
            blhc.task_data,
            {
                "enable_regression_tracking": True,
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            blhc.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "build log hardening check",
                "step": "blhc",
            },
        )
        self.assertQuerySetEqual(
            blhc.dependencies.values_list("task_type", "task_name"),
            [(TaskTypes.WORKER, "sbuild")],
        )
        # No children yet.
        self.assertFalse(blhc.children.exists())

        # Complete sbuild to unblock the blhc workflow.
        assert workflow.parent is not None
        self.simulate_sbuild_workflow_completion(
            workflow.parent.children.get(
                task_type=TaskTypes.WORKFLOW, task_name="sbuild"
            )
        )
        blhc.refresh_from_db()
        self.assertEqual(blhc.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(workflow.parent)
        blhc.refresh_from_db()
        self.assertEqual(blhc.status, WorkRequest.Statuses.RUNNING)
        self.assertTrue(blhc.children.exists())

        ra = workflow.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )
        self.assertEqual(ra.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(
            ra.workflow_data_json,
            {
                "allow_failure": False,
                "allow_dependency_failures": True,
                "display_name": "Regression analysis",
                "step": "regression-analysis",
                "visible": True,
            },
        )
        ra_callbacks = [
            wr.children.get(
                workflow_data_json__step__startswith="regression-analysis"
            )
            for wr in (autopkgtest, sub_autopkgtest, lintian, piuparts, blhc)
        ]
        self.assertQuerySetEqual(
            ra.dependencies.all(), ra_callbacks, ordered=False
        )

    @preserve_db_task_registry()
    def test_populate_architectures_allowed(self) -> None:
        """Populate uses architectures and architectures_allowed."""
        source_artifact = self.source_artifact
        binary_artifact_amd = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="amd64"
            )
        )
        binary_artifact_i386 = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="i386"
            )
        )
        binary_artifact_arm64 = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="arm64"
            )
        )

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [
                    binary_artifact_amd.id,
                    binary_artifact_i386.id,
                    binary_artifact_arm64.id,
                ],
                "architectures": ["amd64", "i386"],
                "architectures_allowlist": ["i386"],
            },
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            autopkgtest.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["i386"],
                "backend": BackendType.AUTO,
                "binary_artifacts": [binary_artifact_i386.id],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "extra_repositories": None,
            },
        )

    @preserve_db_task_registry()
    def test_architectures_deny(self) -> None:
        """Populate uses architectures and architectures_denylist."""
        source_artifact = self.source_artifact
        binary_artifact_amd = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="amd64"
            )
        )
        binary_artifact_i386 = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="i386"
            )
        )
        binary_artifact_arm64 = (
            self.playground.create_minimal_binary_package_artifact(
                architecture="arm64"
            )
        )

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [
                    binary_artifact_amd.id,
                    binary_artifact_i386.id,
                    binary_artifact_arm64.id,
                ],
                "architectures": ["amd64", "i386"],
                "architectures_denylist": ["i386"],
            },
            architectures=["amd64", "i386"],
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            autopkgtest.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "binary_artifacts": [binary_artifact_amd.id],
                "codename": "bookworm",
                "enable_regression_tracking": False,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "source_artifact": source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "extra_repositories": None,
            },
        )

    @preserve_db_task_registry()
    def test_populate_subworkflow_configuration(self) -> None:
        """Populate configures sub-workflows according to specified data."""
        self.playground.create_debian_environment()
        source_artifact = self.source_artifact
        binary_artifact = self.binary_artifact

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [binary_artifact.id],
                "autopkgtest_backend": BackendType.INCUS_VM,
                "lintian_backend": BackendType.UNSHARE,
                "lintian_fail_on_severity": LintianFailOnSeverity.PEDANTIC,
                "piuparts_backend": BackendType.INCUS_LXC,
                "piuparts_environment": "debian/match:codename=trixie",
            },
            sbuild_workflow_source_artifact=source_artifact,
        )

        autopkgtest = workflow.children.get(
            task_name="autopkgtest", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(autopkgtest.task_data["backend"], BackendType.INCUS_VM)

        lintian = workflow.children.get(
            task_name="lintian", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(lintian.task_data["backend"], BackendType.UNSHARE)
        self.assertEqual(
            lintian.task_data["fail_on_severity"],
            LintianFailOnSeverity.PEDANTIC,
        )

        piuparts = workflow.children.get(
            task_name="piuparts", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(piuparts.task_data["backend"], BackendType.INCUS_LXC)
        self.assertEqual(
            piuparts.task_data["environment"], "debian/match:codename=trixie"
        )

    @preserve_db_task_registry()
    def test_populate_debdiff(self) -> None:
        """Populate creates a debdiff workflow."""
        self.playground.create_collection(
            name="bookworm",
            category=CollectionCategory.SUITE,
        )
        trixie = self.playground.create_collection(
            name="trixie",
            category=CollectionCategory.SUITE,
        )
        origin_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0.0-1"
        )
        trixie.manager.add_artifact(
            origin_source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        origin_binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0.0-1",
                version="1.0.0-1",
                architecture="amd64",
            )
        )
        trixie.manager.add_artifact(
            origin_binary_artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        source_artifact = self.playground.create_source_artifact(name="hello")
        workflow = self.orchestrate(
            architectures=["amd64"],
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": LookupMultiple(
                    ("internal@collections/name:build-amd64",)
                ),
                "architectures": ["amd64"],
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": False,
                "enable_debdiff": True,
                "vendor": "debian",
                "codename": "bookworm",
                "qa_suite": "trixie@debian:suite",
            },
            sbuild_workflow_source_artifact=source_artifact,
        )
        self.playground.advance_work_request(workflow, mark_running=True)

        assert workflow.parent is not None
        sbuild_workflow = workflow.parent.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        self.playground.advance_work_request(sbuild_workflow, mark_running=True)

        sbuild_worker_amd64 = sbuild_workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKER
        )

        debdiff_workflow = workflow.children.get(
            task_name="debdiff", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            debdiff_workflow.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "codename": "bookworm",
                "original": "trixie@debian:suite",
                "source_artifact": source_artifact.id,
                "vendor": "debian",
            },
        )

        # No children and blocked: it needs the binary artifacts to be able
        # to create the debdiff tasks
        self.assertEqual(debdiff_workflow.children.count(), 0)
        self.assertEqual(debdiff_workflow.status, WorkRequest.Statuses.BLOCKED)
        self.assertQuerySetEqual(
            debdiff_workflow.dependencies.all(), [sbuild_worker_amd64]
        )

        # Create build artifacts and complete sbuild to unblock the debdiff
        # workflow.
        binaries, _ = self.simulate_sbuild_workflow_completion(sbuild_workflow)

        debdiff_workflow.refresh_from_db()
        self.assertEqual(debdiff_workflow.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(workflow.parent)

        debdiff_workflow.refresh_from_db()
        self.assertEqual(debdiff_workflow.status, WorkRequest.Statuses.RUNNING)
        [debdiff_source_work_request, debdiff_binary_work_request] = (
            debdiff_workflow.children.order_by("id")
        )

        self.assertEqual(debdiff_source_work_request.task_name, "debdiff")
        self.assertEqual(
            debdiff_source_work_request.workflow_data_json,
            {
                "display_name": "DebDiff for source package",
                "step": "debdiff-source",
            },
        )
        self.assertEqual(
            debdiff_source_work_request.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bookworm",
                "extra_flags": [],
                "input": {
                    "source_artifacts": [
                        "trixie@debian:suite/name:hello_1.0.0-1",
                        source_artifact.id,
                    ]
                },
            },
        )

        self.assertEqual(
            debdiff_binary_work_request.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bookworm",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        ["trixie@debian:suite/name:hello_1.0.0-1_amd64"],
                        [binaries["amd64"][0].id],
                    ]
                },
            },
        )

    @preserve_db_task_registry()
    def test_populate_blhc(self) -> None:
        """Populate enables blhc."""
        architectures = ["amd64"]
        source_artifact = self.source_artifact
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        assert isinstance(sid.manager, DebianSuiteManager)

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "architectures": ["amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "enable_blhc": True,
            },
            architectures=architectures,
            sbuild_workflow_source_artifact=source_artifact,
        )

        assert workflow.parent is not None
        sbuild_workflow = workflow.parent.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        self.playground.advance_work_request(sbuild_workflow, mark_running=True)

        blhc_workflow = workflow.children.get(
            task_name="blhc", task_type=TaskTypes.WORKFLOW
        )

        # No children and blocked: it needs the binary artifacts first.
        self.assertFalse(blhc_workflow.children.exists())
        self.assertEqual(blhc_workflow.status, WorkRequest.Statuses.BLOCKED)
        self.assertQuerySetEqual(
            blhc_workflow.dependencies.values_list("task_type", "task_name"),
            [(TaskTypes.WORKER, "sbuild")],
        )

        # Create build artifacts and complete sbuild to unblock the blhc
        # workflow.
        _, build_logs = self.simulate_sbuild_workflow_completion(
            sbuild_workflow
        )

        blhc_workflow.refresh_from_db()
        self.assertEqual(blhc_workflow.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(workflow.parent)

        blhc_workflow.refresh_from_db()
        self.assertEqual(blhc_workflow.children.count(), 1)
        self.assertEqual(blhc_workflow.status, WorkRequest.Statuses.RUNNING)

        [blhc_work_request] = blhc_workflow.children.all()

        self.assertEqual(blhc_work_request.task_name, "blhc")
        self.assertEqual(
            blhc_work_request.workflow_data_json,
            {
                "display_name": "build log hardening check for amd64",
                "step": "blhc-amd64",
            },
        )
        self.assertEqual(
            blhc_work_request.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bookworm",
                "extra_flags": [],
                "input": {"artifact": f"{build_logs['amd64'].id}@artifacts"},
            },
        )

    def test_summarize_status(self) -> None:
        for statuses, expected_status in (
            (
                {
                    RegressionAnalysisStatus.REGRESSION,
                    RegressionAnalysisStatus.ERROR,
                    RegressionAnalysisStatus.IMPROVEMENT,
                    RegressionAnalysisStatus.STABLE,
                    RegressionAnalysisStatus.NO_RESULT,
                },
                RegressionAnalysisStatus.REGRESSION,
            ),
            (
                {
                    RegressionAnalysisStatus.ERROR,
                    RegressionAnalysisStatus.IMPROVEMENT,
                    RegressionAnalysisStatus.STABLE,
                    RegressionAnalysisStatus.NO_RESULT,
                },
                RegressionAnalysisStatus.ERROR,
            ),
            (
                {
                    RegressionAnalysisStatus.IMPROVEMENT,
                    RegressionAnalysisStatus.STABLE,
                    RegressionAnalysisStatus.NO_RESULT,
                },
                RegressionAnalysisStatus.IMPROVEMENT,
            ),
            (
                {
                    RegressionAnalysisStatus.STABLE,
                    RegressionAnalysisStatus.NO_RESULT,
                },
                RegressionAnalysisStatus.STABLE,
            ),
            (
                {RegressionAnalysisStatus.NO_RESULT},
                RegressionAnalysisStatus.NO_RESULT,
            ),
            (set(), RegressionAnalysisStatus.NO_RESULT),
        ):
            with self.subTest(statuses=repr(statuses)):
                self.assertEqual(
                    QAWorkflow._summarize_status(statuses), expected_status
                )

    def _create_autopkgtest_artifact(
        self,
        statuses: dict[str, DebianAutopkgtestResultStatus],
        source_artifact: Artifact,
        work_request: WorkRequest,
        *,
        architecture: str = "amd64",
        distribution: str = "sid",
    ) -> Artifact:
        """Create autopkgtest artifact for regression analysis tests."""
        source_data = DebianSourcePackage(**source_artifact.data)
        source_url = f"http://example.com/{source_artifact.get_absolute_url()}"
        return self.playground.create_artifact(
            category=ArtifactCategory.AUTOPKGTEST,
            data=DebianAutopkgtest(
                results={
                    name: DebianAutopkgtestResult(status=status)
                    for name, status in statuses.items()
                },
                cmdline="unused",
                source_package=DebianAutopkgtestSource(
                    name=get_source_package_name(source_data),
                    version=get_source_package_version(source_data),
                    url=source_url,
                ),
                architecture=architecture,
                distribution=distribution,
            ),
            work_request=work_request,
        )[0]

    def _create_lintian_artifact(
        self,
        tags_count_by_severity: dict[DebianLintianSeverity, int],
        tags_found: list[str],
        work_request: WorkRequest,
        *,
        architecture: str = "amd64",
        distribution: str = "sid",
    ) -> Artifact:
        """Create lintian artifact for regression analysis tests."""
        return self.playground.create_artifact(
            category=ArtifactCategory.LINTIAN,
            data=DebianLintian(
                architecture=architecture,
                summary=DebianLintianSummary(
                    tags_count_by_severity=tags_count_by_severity,
                    package_filename={},
                    tags_found=tags_found,
                    overridden_tags_found=[],
                    lintian_version="2.122.0",
                    distribution=distribution,
                ),
            ),
            work_request=work_request,
        )[0]

    def _prepare_test_callback_regression_analysis(
        self,
    ) -> tuple[WorkRequest, dict[str, Artifact], Artifact]:
        reference_source_artifact = self.source_artifact
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        [new_binary_upload_artifact] = self.playground.create_upload_artifacts(
            source=False,
            src_name="hello",
            version="1.0-1",
            binary=True,
            binaries=[("hello", "amd64")],
        ).binaries
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            name="sid", category=CollectionCategory.QA_RESULTS
        )

        reference_autopkgtest_task = self.playground.create_worker_task(
            task_name="autopkgtest",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        reference_autopkgtest_artifact = self._create_autopkgtest_artifact(
            {"upstream": DebianAutopkgtestResultStatus.PASS},
            reference_source_artifact,
            reference_autopkgtest_task,
        )
        sid_qa_results.manager.add_artifact(
            reference_autopkgtest_artifact,
            user=self.playground.get_default_user(),
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(timezone.now().timestamp()),
                "work_request_id": reference_autopkgtest_task.id,
            },
        )

        reference_lintian_task = self.playground.create_worker_task(
            task_name="lintian",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        reference_lintian_artifact = self._create_lintian_artifact(
            {}, [], reference_lintian_task
        )
        self.playground.create_artifact_relation(
            reference_lintian_artifact, reference_source_artifact
        )
        sid_qa_results.manager.add_artifact(
            reference_lintian_artifact,
            user=self.playground.get_default_user(),
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(timezone.now().timestamp()),
                "work_request_id": reference_lintian_task.id,
            },
        )

        workflow = self.orchestrate(
            extra_data={
                "reference_prefix": "reference-qa-result|",
                "source_artifact": new_source_artifact.id,
                "binary_artifacts": ["internal@collections/name:build-amd64"],
                "reference_source_artifact": reference_source_artifact.id,
                "architectures": ["amd64"],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": True,
                # Simplify this test by only enabling autopkgtest and lintian.
                "enable_autopkgtest": True,
                "enable_lintian": True,
                "enable_piuparts": False,
                "enable_blhc": False,
            },
            architectures=["amd64"],
            sbuild_workflow_source_artifact=new_source_artifact,
            internal_collection_artifacts=[
                (new_binary_upload_artifact, {"name": "build-amd64"})
            ],
        )
        return (
            workflow,
            {
                "autopkgtest:hello:amd64": reference_autopkgtest_artifact,
                "lintian:hello:amd64": reference_lintian_artifact,
            },
            new_source_artifact,
        )

    @preserve_db_task_registry()
    def test_callback_regression_analysis_success(self) -> None:
        """The collation callback succeeds if there were no regressions."""
        (workflow, reference_results, new_source_artifact) = (
            self._prepare_test_callback_regression_analysis()
        )
        assert workflow.parent is not None
        new_results: dict[str, Artifact] = {}
        analysis: dict[
            str, tuple[RegressionAnalysisStatus, dict[str, Any] | None]
        ] = {
            "": (RegressionAnalysisStatus.NO_RESULT, None),
            "autopkgtest:hello:amd64": (
                RegressionAnalysisStatus.NO_RESULT,
                {"upstream": "stable"},
            ),
            "lintian:hello:amd64": (
                RegressionAnalysisStatus.NO_RESULT,
                {"added_tags": [], "removed_tags": []},
            ),
        }

        # Add an extra sub-workflow with no regression-analysis callbacks,
        # to make sure that doesn't cause a problem.
        noop = workflow.create_child_workflow(task_name="noop")

        sbuild = workflow.parent.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        autopkgtest = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        autopkgtest_ra = autopkgtest.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )
        lintian = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        lintian_ra = lintian.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )
        ra = workflow.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )

        with self.schedule_and_run() as work_requests:
            self.assertCountEqual(
                work_requests, [autopkgtest_ra, lintian_ra, ra]
            )

        autopkgtest_ra.refresh_from_db()
        self.assertEqual(autopkgtest_ra.status, WorkRequest.Statuses.BLOCKED)
        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.BLOCKED)
        workflow.refresh_from_db()
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )

        # Complete sbuild and autopkgtest.  The top-level analysis is updated.
        self.playground.advance_work_request(
            sbuild.children.get(task_type=TaskTypes.WORKER, task_name="sbuild"),
            result=WorkRequest.Results.SUCCESS,
        )
        autopkgtest_task = autopkgtest.children.get(
            task_type=TaskTypes.WORKER, task_name="autopkgtest"
        )
        new_results["autopkgtest:hello:amd64"] = (
            self._create_autopkgtest_artifact(
                {"upstream": DebianAutopkgtestResultStatus.PASS},
                new_source_artifact,
                autopkgtest_task,
            )
        )
        self.playground.advance_work_request(
            autopkgtest_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        analysis[""] = (RegressionAnalysisStatus.STABLE, None)
        analysis["autopkgtest:hello:amd64"] = (
            RegressionAnalysisStatus.STABLE,
            {"upstream": "stable"},
        )

        self.schedule_and_run_workflow_callback(autopkgtest)
        self.schedule_and_run_workflow_callback(workflow)

        autopkgtest.refresh_from_db()
        self.assertEqual(autopkgtest.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(autopkgtest.result, WorkRequest.Results.SUCCESS)
        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.BLOCKED)
        workflow.refresh_from_db()
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )

        # Complete lintian.  The top-level analysis is updated, and we're done.
        lintian_task = lintian.children.get(
            task_type=TaskTypes.WORKER, task_name="lintian"
        )
        new_results["lintian:hello:amd64"] = self._create_lintian_artifact(
            {}, [], lintian_task
        )
        self.playground.create_artifact_relation(
            new_results["lintian:hello:amd64"], new_source_artifact
        )
        self.playground.advance_work_request(
            lintian_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        analysis["lintian:hello:amd64"] = (
            RegressionAnalysisStatus.STABLE,
            {"added_tags": [], "removed_tags": []},
        )
        self.playground.advance_work_request(
            noop, result=WorkRequest.Results.SUCCESS
        )

        self.schedule_and_run_workflow_callback(lintian)
        self.schedule_and_run_workflow_callback(workflow)

        lintian.refresh_from_db()
        self.assertEqual(lintian.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(lintian.result, WorkRequest.Results.SUCCESS)
        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(ra.result, WorkRequest.Results.SUCCESS)
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(workflow.result, WorkRequest.Results.SUCCESS)
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )

    @preserve_db_task_registry()
    def test_callback_regression_analysis_failure(self) -> None:
        """The collation callback fails if there were regressions."""
        (workflow, reference_results, new_source_artifact) = (
            self._prepare_test_callback_regression_analysis()
        )
        assert workflow.parent is not None
        new_results: dict[str, Artifact] = {}
        analysis: dict[
            str, tuple[RegressionAnalysisStatus, dict[str, Any] | None]
        ] = {
            "": (RegressionAnalysisStatus.NO_RESULT, None),
            "autopkgtest:hello:amd64": (
                RegressionAnalysisStatus.NO_RESULT,
                {"upstream": "stable"},
            ),
            "lintian:hello:amd64": (
                RegressionAnalysisStatus.NO_RESULT,
                {"added_tags": [], "removed_tags": []},
            ),
        }

        sbuild = workflow.parent.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        autopkgtest = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        autopkgtest_ra = autopkgtest.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )
        lintian = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        lintian_ra = lintian.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )
        ra = workflow.children.workflow_callbacks().get(
            workflow_data_json__step="regression-analysis"
        )

        with self.schedule_and_run() as work_requests:
            self.assertCountEqual(
                work_requests, [autopkgtest_ra, lintian_ra, ra]
            )

        autopkgtest_ra.refresh_from_db()
        self.assertEqual(autopkgtest_ra.status, WorkRequest.Statuses.BLOCKED)
        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.BLOCKED)
        workflow.refresh_from_db()
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )

        # Complete sbuild and autopkgtest.  The top-level analysis is updated.
        self.playground.advance_work_request(
            sbuild.children.get(task_type=TaskTypes.WORKER, task_name="sbuild"),
            result=WorkRequest.Results.SUCCESS,
        )
        autopkgtest_task = autopkgtest.children.get(
            task_type=TaskTypes.WORKER, task_name="autopkgtest"
        )
        new_results["autopkgtest:hello:amd64"] = (
            self._create_autopkgtest_artifact(
                {"upstream": DebianAutopkgtestResultStatus.FAIL},
                new_source_artifact,
                autopkgtest_task,
            )
        )
        self.playground.advance_work_request(
            autopkgtest_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        analysis[""] = (RegressionAnalysisStatus.REGRESSION, None)
        analysis["autopkgtest:hello:amd64"] = (
            RegressionAnalysisStatus.REGRESSION,
            {"upstream": "regression"},
        )

        self.schedule_and_run_workflow_callback(autopkgtest)
        self.schedule_and_run_workflow_callback(workflow)

        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.BLOCKED)
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.RUNNING)
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )

        # Complete lintian.  The top-level analysis is updated, and we're done.
        lintian_task = lintian.children.get(
            task_type=TaskTypes.WORKER, task_name="lintian"
        )
        new_results["lintian:hello:amd64"] = self._create_lintian_artifact(
            {DebianLintianSeverity.ERROR: 1}, ["test"], lintian_task
        )
        self.playground.create_artifact_relation(
            new_results["lintian:hello:amd64"], new_source_artifact
        )
        self.playground.advance_work_request(
            lintian_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        analysis["lintian:hello:amd64"] = (
            RegressionAnalysisStatus.REGRESSION,
            {"added_tags": ["test"], "removed_tags": []},
        )

        self.schedule_and_run_workflow_callback(lintian)
        self.schedule_and_run_workflow_callback(workflow)

        ra.refresh_from_db()
        self.assertEqual(ra.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(ra.result, WorkRequest.Results.FAILURE)
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(workflow.result, WorkRequest.Results.FAILURE)
        self.assert_regression_analysis(
            workflow, "1.0-1", reference_results, "1.0-2", new_results, analysis
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the ``update_suites`` workflow."""

import logging
from typing import override

from django.utils import timezone

import debusine.tasks.tags as ttags
from debusine.artifacts.models import CollectionCategory, TaskTypes
from debusine.assets import KeyPurpose
from debusine.assets.models import SigningKeyData
from debusine.db.models import (
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.server.workflows import (
    UpdateSuitesWorkflow,
    WorkflowValidationError,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks.models import BaseDynamicTaskData


class UpdateSuitesWorkflowTests(WorkflowTestCase[UpdateSuitesWorkflow]):
    """Unit tests for :py:class:`UpdateSuitesWorkflow`."""

    scenario = scenarios.DefaultContext()

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.playground.add_user(
            cls.scenario.workspace_owners, cls.scenario.user
        )

    def create_binary_package_item(self, suite: Collection) -> CollectionItem:
        """Create a minimal debian:binary-package item in the suite."""
        return CollectionItem.objects.create_from_artifact(
            self.playground.create_minimal_binary_package_artifact(),
            parent_collection=suite,
            name="hello_1.0_amd64",
            data={},
            created_by_user=self.playground.get_default_user(),
        )

    def create_release_item(self, suite: Collection) -> CollectionItem:
        """Create a minimal debian:repository-index item in the suite."""
        return CollectionItem.objects.create_from_artifact(
            self.playground.create_repository_index("Release"),
            parent_collection=suite,
            name="index:Release",
            data={"path": "Release"},
            created_by_user=self.playground.get_default_user(),
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.GROUP_PREFIX + "debusine:System:owner"}
        )

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task, strip_task_inputs=True),
            BaseDynamicTaskData(),
        )

    def test_validate_input(self) -> None:
        self.task.validate_input()

    def test_validate_input_not_workspace_owner(self) -> None:
        """Only a workspace owner may run this workflow."""
        self.scenario.workspace_owners.users.remove(self.scenario.user)
        workflow = self.playground.create_workflow(
            task_name="update_suites",
            workspace=self.scenario.workspace,
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            'Only a workspace owner may run the "update_suites" workflow',
        ):
            self.get_workflow(workflow).validate_input()

    def test_find_stale_suites(self) -> None:
        """Suites with changes after their latest Release file are selected."""
        workspaces = [
            self.playground.create_workspace(name=name, public=True)
            for name in ("one", "two")
        ]
        for workspace in workspaces:
            self.playground.create_group_role(
                workspace, Workspace.Roles.OWNER, [self.scenario.user]
            )
        suites = {
            name: self.playground.create_collection(
                name, CollectionCategory.SUITE, workspace=workspace
            )
            for name, workspace in (
                ("empty-unchanged", workspaces[0]),
                ("empty-item-created", workspaces[0]),
                ("release-unchanged", workspaces[0]),
                ("release-item-created", workspaces[0]),
                ("release-item-removed", workspaces[0]),
                ("different-workspace", workspaces[1]),
            )
        }
        self.create_binary_package_item(suites["release-unchanged"])
        release_item_removed_binary = self.create_binary_package_item(
            suites["release-item-removed"]
        )
        self.create_release_item(suites["release-unchanged"])
        self.create_release_item(suites["release-item-created"])
        self.create_release_item(suites["release-item-removed"])
        self.create_binary_package_item(suites["empty-item-created"])
        self.create_binary_package_item(suites["release-item-created"])
        release_item_removed_binary.removed_at = timezone.now()
        release_item_removed_binary.save()
        self.create_binary_package_item(suites["different-workspace"])
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspaces[0]
        )

        self.assertQuerySetEqual(
            self.get_workflow(workflow)._find_stale_suites(),
            [
                suites[name]
                for name in sorted(
                    [
                        "empty-item-created",
                        "release-item-created",
                        "release-item-removed",
                    ]
                )
            ],
        )

    def test_find_stale_suites_force_basic_indexes(self) -> None:
        """``force_basic_indexes`` selects all suites in the workspace."""
        workspaces = [
            self.playground.create_workspace(name=name, public=True)
            for name in ("one", "two")
        ]
        for workspace in workspaces:
            self.playground.create_group_role(
                workspace, Workspace.Roles.OWNER, [self.scenario.user]
            )
        suites = {
            name: self.playground.create_collection(
                name, CollectionCategory.SUITE, workspace=workspace
            )
            for name, workspace in (
                ("one", workspaces[0]),
                ("two", workspaces[0]),
                ("three", workspaces[0]),
                ("four", workspaces[1]),
            )
        }
        workflow = self.playground.create_workflow(
            task_name="update_suites",
            task_data={"force_basic_indexes": True},
            workspace=workspaces[0],
        )

        self.assertQuerySetEqual(
            self.get_workflow(workflow)._find_stale_suites(),
            [suites[name] for name in sorted(["one", "two", "three"])],
        )

    def test_find_stale_suites_only_suites(self) -> None:
        """``only_suites`` limits to suites with the given names."""
        suites = {
            name: self.playground.create_collection(
                name, CollectionCategory.SUITE
            )
            for name in ("one", "two", "three")
        }
        for suite in suites.values():
            self.create_binary_package_item(suite)
        workflow = self.playground.create_workflow(
            task_name="update_suites", task_data={"only_suites": ["one", "two"]}
        )

        self.assertQuerySetEqual(
            self.get_workflow(workflow)._find_stale_suites(),
            [suites[name] for name in sorted(["one", "two"])],
        )

    def test_populate(self) -> None:
        """``populate`` creates child work requests for all stale suites."""
        workspace = self.scenario.workspace
        archive = self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        suites = [
            self.playground.create_collection(name, CollectionCategory.SUITE)
            for name in ("one", "two")
        ]
        for suite in suites:
            self.create_binary_package_item(suite)
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspace
        )

        workflow.set_current()
        self.assertTrue(orchestrate_workflow(workflow))

        self.assertQuerySetEqual(
            workflow.children.values_list(
                "task_type",
                "task_name",
                "task_data",
                "workflow_data_json",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.SIGNING,
                    "generatekey",
                    {
                        "purpose": KeyPurpose.OPENPGP_REPOSITORY,
                        "description": f"Archive signing key for {workspace}",
                    },
                    {
                        "display_name": f"Generate signing key for {workspace}",
                        "step": "generate-signing-key",
                    },
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.INTERNAL,
                    "workflow",
                    {},
                    {
                        "display_name": "Record generated key",
                        "step": "generated-key",
                    },
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.SIGNING,
                    "generatekey",
                ),
                (
                    TaskTypes.WORKFLOW,
                    "update_suite",
                    {"suite_collection": suites[0].id},
                    {"display_name": "Update one", "step": "update-one"},
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.INTERNAL,
                    "workflow",
                ),
                (
                    TaskTypes.WORKFLOW,
                    "update_suite",
                    {"suite_collection": suites[1].id},
                    {"display_name": "Update two", "step": "update-two"},
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.INTERNAL,
                    "workflow",
                ),
            ],
        )

        # Completing each step allows the next to proceed.
        generate_key = workflow.children.get(
            task_type=TaskTypes.SIGNING, task_name="generatekey"
        )
        signing_key = self.playground.create_signing_key_asset(
            purpose=KeyPurpose.OPENPGP_REPOSITORY,
            fingerprint="ABC123",
            public_key="Fake public key",
            description=f"Generate signing key for {workspace}",
            workspace=workspace,
            created_by_work_request=generate_key,
        )
        self.playground.advance_work_request(
            generate_key, result=WorkRequest.Results.SUCCESS
        )

        generated_key = workflow.children.workflow_callbacks().get()
        self.assertEqual(generated_key.status, WorkRequest.Statuses.PENDING)

        self.assertTrue(orchestrate_workflow(workflow))

        generated_key.refresh_from_db()
        self.assertEqual(generated_key.status, WorkRequest.Statuses.COMPLETED)
        archive.refresh_from_db()
        self.assertTrue(signing_key.usage.filter(workspace=workspace).exists())
        signing_key_data = signing_key.data_model
        assert isinstance(signing_key_data, SigningKeyData)
        self.assertEqual(
            archive.data, {"signing_keys": [signing_key_data.fingerprint]}
        )
        update_suite_children = workflow.children.filter(
            task_type=TaskTypes.WORKFLOW, task_name="update_suite"
        )
        self.assertQuerySetEqual(
            update_suite_children.values_list("task_data", "status"),
            [
                (
                    {"suite_collection": suite.id},
                    WorkRequest.Statuses.RUNNING,
                )
                for suite in suites
            ],
        )
        for child in update_suite_children:
            self.assertTrue(child.children.exists())

    def test_populate_no_archive(self) -> None:
        """``populate`` skips the case where the workspace has no archive."""
        workspace = self.scenario.workspace
        workspace.collections.filter(
            category=CollectionCategory.ARCHIVE
        ).delete()
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspace
        )

        with self.assertLogsContains(
            f"Workspace {workspace} has no archive; not updating suites.",
            logger="debusine.server.workflows.update_suites",
            level=logging.INFO,
        ):
            self.assertTrue(orchestrate_workflow(workflow))

        self.assertFalse(workflow.children.exists())

    def test_populate_not_exported(self) -> None:
        """``populate`` skips suites that are marked as not to be exported."""
        workspace = self.scenario.workspace
        self.playground.create_singleton_collection(CollectionCategory.ARCHIVE)
        suites = [
            self.playground.create_collection(
                name, CollectionCategory.SUITE, data=data
            )
            for name, data in (("one", {}), ("two", {"exported": False}))
        ]
        for suite in suites:
            self.create_binary_package_item(suite)
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspace
        )

        workflow.set_current()
        orchestrate_workflow(workflow)

        self.assertQuerySetEqual(
            workflow.children.filter(
                task_type=TaskTypes.WORKFLOW, task_name="update_suite"
            ).values_list("task_data__suite_collection", flat=True),
            [suites[0].id],
        )

    def test_populate_already_exists(self) -> None:
        """``populate`` skips work requests that already exist."""
        workspace = self.scenario.workspace
        self.playground.create_singleton_collection(CollectionCategory.ARCHIVE)
        suites = [
            self.playground.create_collection(name, CollectionCategory.SUITE)
            for name in ("one", "two")
        ]
        self.create_binary_package_item(suites[0])
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspace, mark_running=True
        )

        self.assertTrue(orchestrate_workflow(workflow))

        expected_children = [
            (
                TaskTypes.SIGNING,
                "generatekey",
                {
                    "purpose": KeyPurpose.OPENPGP_REPOSITORY,
                    "description": f"Archive signing key for {workspace}",
                },
                {
                    "display_name": f"Generate signing key for {workspace}",
                    "step": "generate-signing-key",
                },
                WorkRequest.Statuses.PENDING,
                None,
                None,
            ),
            (
                TaskTypes.INTERNAL,
                "workflow",
                {},
                {
                    "display_name": "Record generated key",
                    "step": "generated-key",
                },
                WorkRequest.Statuses.BLOCKED,
                TaskTypes.SIGNING,
                "generatekey",
            ),
            (
                TaskTypes.WORKFLOW,
                "update_suite",
                {"suite_collection": suites[0].id},
                {"display_name": "Update one", "step": "update-one"},
                WorkRequest.Statuses.BLOCKED,
                TaskTypes.INTERNAL,
                "workflow",
            ),
        ]
        self.assertQuerySetEqual(
            workflow.children.values_list(
                "task_type",
                "task_name",
                "task_data",
                "workflow_data_json",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            expected_children,
        )

        self.create_binary_package_item(suites[1])

        self.assertTrue(orchestrate_workflow(workflow))

        expected_children.append(
            (
                TaskTypes.WORKFLOW,
                "update_suite",
                {"suite_collection": suites[1].id},
                {"display_name": "Update two", "step": "update-two"},
                WorkRequest.Statuses.BLOCKED,
                TaskTypes.INTERNAL,
                "workflow",
            )
        )
        self.assertQuerySetEqual(
            workflow.children.values_list(
                "task_type",
                "task_name",
                "task_data",
                "workflow_data_json",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            expected_children,
        )

    def test_populate_signing_keys_already_exist(self) -> None:
        """``populate`` only generates signing keys if they don't exist."""
        workspace = self.scenario.workspace
        archive = self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        archive.data = {"signing_keys": ["ABC123"]}
        archive.save()
        suites = [
            self.playground.create_collection(name, CollectionCategory.SUITE)
            for name in ("one", "two")
        ]
        for suite in suites:
            self.create_binary_package_item(suite)
        workflow = self.playground.create_workflow(
            task_name="update_suites", workspace=workspace, mark_running=True
        )

        self.assertTrue(orchestrate_workflow(workflow))

        self.assertQuerySetEqual(
            workflow.children.values_list(
                "task_type",
                "task_name",
                "task_data",
                "workflow_data_json",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKFLOW,
                    "update_suite",
                    {"suite_collection": suite.id},
                    {
                        "display_name": f"Update {suite.name}",
                        "step": f"update-{suite.name}",
                    },
                    WorkRequest.Statuses.RUNNING,
                    None,
                    None,
                )
                for suite in suites
            ],
        )
        for child in workflow.children.all():
            self.assertTrue(child.children.exists())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the SbuildWorkflow class."""

from typing import Any, ClassVar, override

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.db.models import (
    Artifact,
    CollectionItem,
    WorkRequest,
    WorkflowTemplate,
    default_workspace,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections.tests.utils import CollectionTestMixin
from debusine.server.workflows import (
    SbuildWorkflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    SbuildWorkflowData,
    SbuildWorkflowDynamicData,
)
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    ActionTypes,
    ActionUpdateCollectionWithArtifacts,
    ActionUpdateCollectionWithData,
    BackendType,
    InputArtifactSingle,
    SbuildInput,
)
from debusine.test.test_utils import create_system_tarball_data


class SbuildWorkflowTests(
    CollectionTestMixin, WorkflowTestCase[SbuildWorkflow]
):
    """Unit tests for SbuildWorkflow class."""

    SAMPLE_TASK_DATA = {
        "target_distribution": "debian:bookworm",
        "architectures": ["all", "amd64"],
    }

    source_artifact: ClassVar[Artifact]
    template: WorkflowTemplate

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = (
            cls.playground.create_minimal_source_package_artifact(
                name="hello",
                version="1.0-1",
                dsc_fields={"Architecture": "all"},
            )
        )
        cls.playground.create_debian_environment(
            codename="bookworm", architecture="amd64"
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["input"] = {"source_artifact": self.source_artifact.id}
        self.template = self.create_template("sbuild")

    def test_create_orchestrator(self) -> None:
        """Test instantiating a SbuildWorkflow."""
        task_data: dict[str, Any] = {
            "input": {
                "source_artifact": self.source_artifact.id,
            },
            "target_distribution": "debian:bookworm",
            "architectures": ["all", "amd64"],
        }
        wr = self.playground.create_workflow(
            task_name="sbuild",
            task_data=task_data,
            workspace=default_workspace(),
        )
        w = self.get_workflow(wr)
        self.assertEqual(w.data.input.source_artifact, self.source_artifact.id)
        self.assertEqual(w.data.target_distribution, "debian:bookworm")
        self.assertEqual(w.data.backend, BackendType.UNSHARE)
        self.assertEqual(w.data.architectures, ["all", "amd64"])
        self.assertIsNone(w.data.extra_repositories)
        self.assertIsNone(w.data.binnmu)

        task_data["binnmu"] = {
            "changelog": "Rebuild for 64bit time_t",
            "suffix": "+b1",
        }
        task_data["extra_repositories"] = [
            {
                "url": "http://example.net/foo",
                "suite": "bookworm",
                "components": ["main", "non-free"],
            },
        ]
        wr = self.playground.create_workflow(
            task_name="sbuild",
            task_data=task_data,
            workspace=default_workspace(),
        )
        w = self.get_workflow(wr)
        assert w.data.binnmu is not None
        self.assertEqual(w.data.binnmu.suffix, "+b1")
        assert w.data.extra_repositories is not None
        self.assertEqual(w.data.extra_repositories[0].suite, "bookworm")

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data includes a parameter summary."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SbuildWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=self.source_artifact.id,
            ),
        )

    def test_get_input_artifacts(self) -> None:
        wf = self.task
        self.assertIsNone(wf.dynamic_data)
        self.assertEqual(
            wf.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=wf.data.input.source_artifact,
                    label="input.source_artifact",
                    artifact_id=None,
                )
            ],
        )

        wf.dynamic_data = wf.compute_dynamic_data()

        self.assertIsNotNone(wf.dynamic_data)
        self.assertEqual(
            wf.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=wf.data.input.source_artifact,
                    label="input.source_artifact",
                    artifact_id=self.source_artifact.id,
                )
            ],
        )

    def test_validate_input(self) -> None:
        """Test that validate_input passes a valid case."""
        tarball_item = self.playground.create_debian_environment(
            variant="sbuild"
        )

        source = self.source_artifact

        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {
                    "source_artifact": source.pk,
                },
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64"],
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)

        o = self.get_workflow(wr)
        self.assertEqual(
            workflow_utils.source_package(
                o, configuration_key="input.source_artifact"
            ),
            source,
        )
        self.assertEqual(o.architectures, {"all"})
        self.assertEqual(o._get_environment("amd64"), tarball_item.artifact)
        o.validate_input()

        # environment_variant
        self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(
                codename="bookworm", variant="buildd"
            ),
        )
        # add with explicit :variant
        tarball_item = self.playground.create_debian_environment(
            environment=tarball_item.artifact,
            variables={"variant": "buildd", "backend": BackendType.UNSHARE},
        )

        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {
                    "source_artifact": source.pk,
                },
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64"],
                "environment_variant": "buildd",
            },
        )
        o = self.get_workflow(wr)
        self.assertEqual(o._get_environment("amd64"), tarball_item.artifact)
        o.validate_input()

    def test_computation_of_architecture_intersection(self) -> None:
        """Test computation of architectures to build."""
        dsc_archs = {
            "any": {"amd64", "arm64"},
            "all": {
                "all",
            },
            "any all": {"all", "amd64", "arm64"},
            "arm64": {
                "arm64",
            },
            "linux-any": {"amd64", "arm64"},
            "gnu-any-any": {"amd64", "arm64"},
        }
        for dsc_arch, expected_archs in dsc_archs.items():
            source = self.create_source_package(
                "hello", "1.0", dsc_fields={"Architecture": dsc_arch}
            )

            for wr_archs in (["amd64"], ["all"], ["arm64", "all"]):
                with self.subTest("", dsc_arch=dsc_arch, wr_archs=wr_archs):
                    o = self.get_workflow(
                        self.playground.create_workflow(
                            self.template,
                            task_data={
                                "input": {
                                    "source_artifact": source.pk,
                                },
                                "target_distribution": "debian:bookworm",
                                "architectures": wr_archs,
                            },
                            validate=False,
                        ),
                    )
                    archs = expected_archs.intersection(wr_archs)
                    if archs:
                        self.assertSetEqual(o.architectures, archs)
                    else:
                        with self.assertRaisesRegex(
                            WorkflowValidationError,
                            "None of the workflow architectures are supported"
                            " by this package",
                        ):
                            o.architectures

    def test_create_work_request_unshare(self) -> None:
        """Test creating a Sbuild workflow workrequest with unshare."""
        self.playground.create_debian_environment(
            codename="bookworm", architecture="s390x"
        )
        default_workspace().collections.filter(
            category=CollectionCategory.PACKAGE_BUILD_LOGS
        ).delete()

        # Create a source package
        source = self.create_source_package(
            "hello",
            "1.0",
            dsc_fields={
                "Binary": "hello, hello-dev",
                "Architecture": "all amd64 s390x",
            },
        )

        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {
                    "source_artifact": source.pk,
                },
                "backend": BackendType.UNSHARE,
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64", "s390x"],
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(wr))

        expected_children = [
            (
                "all",
                {
                    "build_architecture": "amd64",
                    "build_components": ["all"],
                    "build_profiles": None,
                },
            ),
            (
                "amd64",
                {
                    "build_architecture": "amd64",
                    "build_components": ["any"],
                    "build_profiles": None,
                },
            ),
            (
                "s390x",
                {
                    "build_architecture": "s390x",
                    "build_components": ["any"],
                    "build_profiles": None,
                },
            ),
        ]

        children = list(WorkRequest.objects.filter(parent=wr).order_by("id"))
        self.assertEqual(len(children), len(expected_children))

        for child, expected_child in zip(children, expected_children):
            expected_arch, expected_data = expected_child
            self.assertEqual(child.status, WorkRequest.Statuses.PENDING)
            self.assertEqual(child.task_type, TaskTypes.WORKER)
            self.assertEqual(child.task_name, "sbuild")
            self.assertEqual(
                child.task_data,
                {
                    'backend': BackendType.UNSHARE,
                    'environment': 'debian/match:codename=bookworm',
                    'input': {'source_artifact': source.pk},
                    'binnmu': None,
                    'extra_repositories': None,
                    'build_dep_resolver': None,
                    'aspcud_criteria': None,
                    **expected_data,
                },
            )
            promise = CollectionItem.objects.get(
                parent_collection=wr.internal_collection,
                child_type=CollectionItem.Types.BARE,
                category=BareDataCategory.PROMISE,
                name=f"build-{expected_arch}",
            )
            self.assertEqual(
                promise.data,
                {
                    "promise_work_request_id": child.id,
                    "promise_workflow_id": wr.id,
                    "promise_category": ArtifactCategory.UPLOAD,
                    "binary_names": ["hello", "hello-dev"],
                    "source_package_name": "hello",
                    "architecture": expected_arch,
                },
            )
            self.assert_work_request_event_reactions(
                child,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        artifact_filters={"category": ArtifactCategory.UPLOAD},
                        collection="internal@collections",
                        name_template=f"build-{expected_arch}",
                        variables={
                            "binary_names": ["hello", "hello-dev"],
                            "source_package_name": "hello",
                            "architecture": expected_arch,
                        },
                    ),
                    ActionUpdateCollectionWithArtifacts(
                        artifact_filters={
                            "category": ArtifactCategory.PACKAGE_BUILD_LOG
                        },
                        collection="internal@collections",
                        name_template=f"buildlog-{expected_arch}",
                        variables={
                            "source_package_name": "hello",
                            "source_package_version": "1.0",
                            "architecture": expected_arch,
                        },
                    ),
                ],
            )
            self.assertQuerySetEqual(child.dependencies.all(), [])
            self.assertEqual(
                child.workflow_data_json,
                {
                    "display_name": f"Build {expected_arch}",
                    "step": f"build-{expected_arch}",
                },
            )

        # populate() idempotence
        orchestrator = self.get_workflow(wr)
        orchestrator.populate()
        children = list(WorkRequest.objects.filter(parent=wr))
        self.assertEqual(len(children), len(expected_children))

    def test_create_work_request_arch_all_build_architecture(self) -> None:
        """Test using a non-default `arch_all_build_architecture`."""
        self.playground.create_debian_environment(
            codename="bookworm", architecture="s390x"
        )
        source = self.create_source_package(
            "hello", "1.0", dsc_fields={"Architecture": "all amd64 s390x"}
        )
        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {
                    "source_artifact": source.pk,
                },
                "backend": BackendType.UNSHARE,
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64", "s390x"],
                "arch_all_build_architecture": "s390x",
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(wr))

        expected_children = [
            (
                "all",
                {
                    "build_architecture": "s390x",
                    "build_components": ["all"],
                    "build_profiles": None,
                },
            ),
            (
                "amd64",
                {
                    "build_architecture": "amd64",
                    "build_components": ["any"],
                    "build_profiles": None,
                },
            ),
            (
                "s390x",
                {
                    "build_architecture": "s390x",
                    "build_components": ["any"],
                    "build_profiles": None,
                },
            ),
        ]

        children = list(WorkRequest.objects.filter(parent=wr).order_by("id"))
        self.assertEqual(len(children), len(expected_children))

    def test_create_work_request_retain_build_logs(self) -> None:
        """Test retaining build logs in a collection."""
        build_logs = default_workspace().get_singleton_collection(
            pc=PermissionContext.for_ui_user(
                self.playground.get_default_user()
            ),
            category=CollectionCategory.PACKAGE_BUILD_LOGS,
        )

        # Create a source package
        source = self.create_source_package(
            "hello",
            "1.0",
            dsc_fields={
                "Binary": "hello, hello-dev",
                "Architecture": "amd64",
            },
        )

        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {"source_artifact": source.pk},
                "backend": BackendType.UNSHARE,
                "target_distribution": "debian:bookworm",
                "architectures": ["amd64"],
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(wr))

        child = WorkRequest.objects.get(parent=wr)
        self.assertEqual(child.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "sbuild")
        self.assertEqual(
            child.task_data,
            {
                "input": {"source_artifact": source.pk},
                "build_architecture": "amd64",
                "build_components": ["any"],
                "build_profiles": None,
                "binnmu": None,
                "extra_repositories": None,
                "build_dep_resolver": None,
                "aspcud_criteria": None,
                "backend": BackendType.UNSHARE,
                "environment": "debian/match:codename=bookworm",
            },
        )
        promise = CollectionItem.objects.get(
            parent_collection=wr.internal_collection,
            child_type=CollectionItem.Types.BARE,
            category=BareDataCategory.PROMISE,
            name="build-amd64",
        )
        self.assertEqual(
            promise.data,
            {
                "promise_work_request_id": child.id,
                "promise_workflow_id": wr.id,
                "promise_category": ArtifactCategory.UPLOAD,
                "binary_names": ["hello", "hello-dev"],
                "architecture": "amd64",
                "source_package_name": "hello",
            },
        )
        self.assert_work_request_event_reactions(
            child,
            on_creation=[
                ActionUpdateCollectionWithData(
                    collection=build_logs.id,
                    category=BareDataCategory.PACKAGE_BUILD_LOG,
                    data={
                        "work_request_id": child.id,
                        "vendor": "debian",
                        "codename": "bookworm",
                        "architecture": "amd64",
                        "srcpkg_name": "hello",
                        "srcpkg_version": "1.0",
                    },
                ),
            ],
            on_unblock=[],
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": ArtifactCategory.UPLOAD},
                    collection="internal@collections",
                    name_template="build-amd64",
                    variables={
                        "binary_names": ["hello", "hello-dev"],
                        "architecture": "amd64",
                        "source_package_name": "hello",
                    },
                ),
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={
                        "category": ArtifactCategory.PACKAGE_BUILD_LOG
                    },
                    collection="internal@collections",
                    name_template="buildlog-amd64",
                    variables={
                        "source_package_name": "hello",
                        "source_package_version": "1.0",
                        "architecture": "amd64",
                    },
                ),
                ActionUpdateCollectionWithArtifacts(
                    collection=build_logs.id,
                    variables={
                        "work_request_id": child.id,
                        "vendor": "debian",
                        "codename": "bookworm",
                        "architecture": "amd64",
                        "srcpkg_name": "hello",
                        "srcpkg_version": "1.0",
                    },
                    artifact_filters={
                        "category": ArtifactCategory.PACKAGE_BUILD_LOG
                    },
                ),
            ],
            on_failure=[],
        )
        self.assertQuerySetEqual(child.dependencies.all(), [])
        self.assertEqual(
            child.workflow_data_json,
            {"display_name": "Build amd64", "step": "build-amd64"},
        )

    def test_create_work_request_retry_delays(self) -> None:
        """Test adding automatic retries with delays."""
        source = self.create_source_package(
            "hello",
            "1.0",
            dsc_fields={
                "Binary": "hello, hello-dev",
                "Architecture": "amd64",
            },
        )
        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {"source_artifact": source.pk},
                "backend": BackendType.UNSHARE,
                "target_distribution": "debian:bookworm",
                "architectures": ["amd64"],
                "retry_delays": ["1h", "1d"],
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(wr))

        child = WorkRequest.objects.get(parent=wr)
        self.assertEqual(child.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "sbuild")
        self.assertEqual(
            child.task_data,
            {
                "input": {"source_artifact": source.pk},
                "build_architecture": "amd64",
                "build_components": ["any"],
                "build_profiles": None,
                "binnmu": None,
                "extra_repositories": None,
                "build_dep_resolver": None,
                "aspcud_criteria": None,
                "backend": BackendType.UNSHARE,
                "environment": "debian/match:codename=bookworm",
            },
        )
        self.assertIn(
            ActionUpdateCollectionWithArtifacts(
                artifact_filters={"category": ArtifactCategory.UPLOAD},
                collection="internal@collections",
                created_at=None,
                name_template="build-amd64",
                variables={
                    "binary_names": ["hello", "hello-dev"],
                    "architecture": "amd64",
                    "source_package_name": "hello",
                },
            ),
            child.event_reactions.on_success,
        )
        self.assertIn(
            {
                "action": ActionTypes.RETRY_WITH_DELAYS,
                "delays": ["1h", "1d"],
            },
            child.event_reactions_json["on_failure"],
        )

    def test_create_work_request_signing_template_names(self) -> None:
        """Test signing template handling."""
        source = self.create_source_package(
            "hello",
            "1.0",
            dsc_fields={
                "Binary": "hello-unsigned",
                "Architecture": "amd64",
            },
        )
        wr = self.playground.create_workflow(
            self.template,
            task_data={
                "input": {"source_artifact": source.pk},
                "backend": BackendType.UNSHARE,
                "target_distribution": "debian:bookworm",
                "architectures": ["amd64"],
                "signing_template_names": {"amd64": ["hello-unsigned"]},
            },
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(wr))

        child = WorkRequest.objects.get(parent=wr)
        self.assertEqual(child.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "sbuild")
        self.assertIn(
            ActionUpdateCollectionWithArtifacts(
                artifact_filters={"category": ArtifactCategory.UPLOAD},
                collection="internal@collections",
                created_at=None,
                name_template="build-amd64",
                variables={
                    "binary_names": ["hello-unsigned"],
                    "architecture": "amd64",
                    "source_package_name": "hello",
                },
            ),
            child.event_reactions.on_success,
        )
        self.assertIn(
            ActionUpdateCollectionWithArtifacts(
                artifact_filters={
                    "category": ArtifactCategory.BINARY_PACKAGE,
                    "data__deb_fields__Package": "hello-unsigned",
                },
                collection="internal@collections",
                created_at=None,
                name_template="signing-template-amd64-hello-unsigned",
                variables={
                    "architecture": "amd64",
                    "binary_package_name": "hello-unsigned",
                },
            ),
            child.event_reactions.on_success,
        )

    def test_source_package_error_handling(self) -> None:
        """Test source_package error handling."""
        invalid_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"^input.source_artifact: unexpected artifact category: "
            r"'debusine:test'. Valid categories: "
            r"\['debian:source-package', 'debian:upload'\]$",
        ):
            self.task_create(input={"source_artifact": invalid_artifact.pk})

        self.validate = False
        source = self.create_source_package(
            "hello", "1.0", dsc_fields={"Architecture": ""}
        )
        task = self.task_create(input={"source_artifact": source.pk})
        with self.assertRaisesRegex(
            WorkflowValidationError, "package architecture list is empty"
        ):
            task.architectures

        source = self.create_source_package(
            "hello", "1.0", dsc_fields={"Architecture": "any-armhf"}
        )
        task = self.task_create(
            input={"source_artifact": source.pk},
            architectures=["kfreebsd-i386"],
        )
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "None of the workflow architectures are supported by this package",
        ):
            task.architectures

        # wildcards support
        source = self.create_source_package(
            "hello", "1.0", dsc_fields={"Architecture": "any-amd64"}
        )
        task = self.task_create(
            input={"source_artifact": source.pk},
            architectures=["i386", "hurd-amd64"],
        )
        self.assertEqual(task.architectures, {"hurd-amd64"})

        # wildcards are for packages, not workflows
        source = self.create_source_package(
            "hello", "1.0", dsc_fields={"Architecture": "amd64"}
        )
        task = self.task_create(
            input={"source_artifact": source.pk}, architectures=["any"]
        )
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "None of the workflow architectures are supported by this package",
        ):
            task.architectures

        task = self.task_create(
            input={"source_artifact": source.pk}, backend=BackendType.QEMU
        )
        self.playground.create_debian_environments_collection()
        with self.assertRaisesRegex(
            WorkflowValidationError, "environment not found for"
        ):
            task._get_environment("any")

        task = self.task_create(
            input={"source_artifact": source.pk}, target_distribution="Ubuntu"
        )
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "target_distribution must be in vendor:codename format",
        ):
            task.validate_input()

    def test_populate_upload(self) -> None:
        """The workflow accepts debian:upload artifacts."""
        upload_artifacts = self.playground.create_upload_artifacts()

        wf = self.playground.create_workflow(
            task_name="sbuild",
            task_data=SbuildWorkflowData(
                input=SbuildInput(
                    source_artifact=upload_artifacts.upload.id,
                ),
                target_distribution="debian:bookworm",
                architectures=["amd64"],
            ),
        )
        self.get_workflow(wf).populate()

        children = list(WorkRequest.objects.filter(parent=wf))
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(
            child.task_data["input"]["source_artifact"],
            f"{upload_artifacts.source.id}@artifacts",
        )

    def test_populate_experimental(self) -> None:
        """The workflow handles overlay distributions."""
        self.playground.create_debian_environment(
            codename="experimental", variant="sbuild"
        )
        wr = self.playground.create_workflow(
            task_name="sbuild",
            task_data=SbuildWorkflowData(
                input=SbuildInput(
                    source_artifact=self.source_artifact.id,
                ),
                target_distribution="debian:experimental",
                arch_all_build_architecture="amd64",
                architectures=["all"],
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(len(child.task_data["extra_repositories"]), 1)
        repo = child.task_data["extra_repositories"][0]
        self.assertEqual(repo["suite"], "experimental")

        self.assertEqual(child.task_data["build_dep_resolver"], "aspcud")
        self.assertEqual(
            child.task_data["aspcud_criteria"],
            (
                "-count(down),-count(changed,APT-Release:=/experimental/),"
                "-removed,-changed,-new"
            ),
        )

    def test_label(self) -> None:
        self.assertEqual(self.task.get_label(), "sbuild")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Implementation of external file backends."""

import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Self, override

import requests
from requests_toolbelt.downloadutils.stream import stream_response_to_file

from debusine.db.models import File, FileInStore, FileStore
from debusine.server.file_backend.interface import (
    FileBackendEntryInterface,
    FileBackendInterface,
)
from debusine.server.file_backend.models import (
    ExternalDebianSuiteBackendConfiguration,
)
from debusine.utils import NotSupportedError, get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class ExternalDebianSuiteFileBackendEntry(
    FileBackendEntryInterface["ExternalDebianSuiteFileBackend", str]
):
    """
    An entry in an external repository.

    External repository identifiers are relative URLs under the archive root
    URL.
    """

    @classmethod
    def _create_from_file(
        cls, backend: "ExternalDebianSuiteFileBackend", fileobj: File
    ) -> Self:
        """Make an entry based on a `File`."""
        file_in_store = FileInStore.objects.get(
            file=fileobj, store=backend.db_store
        )
        relative_url = file_in_store.data["relative_url"]
        assert isinstance(relative_url, str)
        return cls(backend=backend, identifier=relative_url)

    @override
    def __str__(self) -> str:
        """Return a string describing this entry."""
        return self.get_url()

    @override
    def get_local_path(self) -> None:
        """
        Return None: no local path for now.

        This may change if we cache the remote file locally.
        """
        return None

    @override
    @contextmanager
    def get_temporary_local_path(self) -> Generator[Path]:
        """Yield a temporary local path for the file."""
        with tempfile.NamedTemporaryFile(
            prefix="debusine-external-", delete_on_close=False
        ) as temp_file:
            headers = {"User-Agent": get_user_agent()}
            with requests.get(
                self.get_url(),
                stream=True,
                headers=headers,
                timeout=DEFAULT_REQUEST_TIMEOUT,
            ) as r:
                r.raise_for_status()
                stream_response_to_file(r, temp_file)
            temp_file.close()
            yield Path(temp_file.name)

    @override
    def get_url(
        self,
        content_type: str | None = None,  # noqa: ARG002, U100
    ) -> str:
        """
        Return a URL pointing to the content.

        We cannot override the Content-Type for external repositories.
        """
        return self.backend.archive_root_url + self._identifier

    @override
    def get_mtime(self) -> None:  # pragma: no cover
        """Return None: no last-modified times for external files."""
        return None

    @override
    def add(self, local_path: Path) -> None:
        """We cannot add files to external repositories."""
        raise NotImplementedError()

    @override
    def same_contents(self, local_path: Path) -> bool:
        """
        We cannot check file contents in external repositories.

        It would be possible, but there's no need since we cannot add files
        to them.
        """
        raise NotImplementedError()

    @override
    def remove(self) -> None:
        """
        Remove the entry from the underlying storage.

        This may change if we cache the remote file locally.
        """
        raise NotImplementedError()


class ExternalDebianSuiteFileBackend(
    FileBackendInterface[ExternalDebianSuiteBackendConfiguration]
):
    """Mirror an external repository in debusine."""

    def __init__(self, file_store: FileStore) -> None:
        """Initialize instance."""
        super().__init__()
        self.db_store = file_store
        # All files are relative to the archive root
        # https://wiki.debian.org/DebianRepository/Format
        self.archive_root_url = self.configuration.archive_root_url
        # For Collections to grab and parse the appropriate Packages files
        self.suite = self.configuration.suite
        self.components = self.configuration.components

    @override
    def get_entry(self, fileobj: File) -> ExternalDebianSuiteFileBackendEntry:
        return ExternalDebianSuiteFileBackendEntry._create_from_file(
            self, fileobj
        )

    @override
    def remove_file(self, db_file: File) -> None:  # noqa: ARG002, U100
        """
        Remove the file from the FileStore.

        This may change if we cache the remote file locally.
        """
        raise NotSupportedError(
            "ExternalDebianSuiteFileBackend is a read-only file backend"
        )

    @override
    def add_file(
        self,
        local_path: Path,  # noqa: ARG002, U100
        db_file: File | None = None,  # noqa: ARG002, U100
    ) -> File:
        raise NotSupportedError(
            "ExternalDebianSuiteFileBackend is a read-only file backend"
        )

    def add_remote_file(
        self,
        relative_url: str,
        hash_digest: bytes,
        size: int,
    ) -> File:
        """Register remote file in the DB."""
        db_file_kwargs: dict[str, Any] = {
            File.current_hash_algorithm: hash_digest,
            'size': size,
        }
        db_file, _ = File.objects.get_or_create(**db_file_kwargs)
        FileInStore.objects.get_or_create(
            store=self.db_store,
            file=db_file,
            data={'relative_url': relative_url},
        )
        return db_file

    @override
    def list_entries(self) -> Generator[ExternalDebianSuiteFileBackendEntry]:
        # We cannot currently iterate over all files in an external suite
        # (though in principle it should be possible).
        raise NotImplementedError()



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""
Debusine generic integration tests.

Does not test any sbuild related code.
"""

import os
import shutil
import textwrap
import unittest
from pathlib import Path
from typing import override

import lxml.html
import requests
import yaml

from utils.client import Client
from utils.common import Configuration
from utils.server import DebusineServer
from utils.worker import Worker


class IntegrationGenericTests(unittest.TestCase):
    """
    Integration generic tests (excludes sbuild specific tests).

    These tests assume:
    - debusine-server is running
    - debusine-client is correctly configured
    - "sudo -u debusine-server debusine-admin COMMAND" works

    Note that these tests (via Worker) delete /etc/debusine/worker contents,
    stop and start debusine-worker, create new tokens in the debusine-server
    """

    TASK_NAME = 'noop'

    @override
    def setUp(self) -> None:
        # If debusine-server or nginx was launched just before the
        # integration-tests.py is launched the debusine-server might not be
        # yet available. Let's wait for the debusine-server to be
        # reachable if it's not ready yet
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )

        self._worker = Worker()

    def connect_worker(self) -> None:
        """
        Connect the worker to the debusine-server.

        Steps:
        - set up worker (it fetches a disabled token file)
        - enable the token
        - wait for the worker to be connected
        """
        self._worker.set_up()

        # Worker re-connect automatically
        self.assertTrue(
            DebusineServer.wait_for_worker_connected(Worker.read_token())
        )

    def test_secret_key_file_not_readable_fails(self) -> None:
        """Key file cannot be read: the server does not start."""
        key_file = "/var/lib/debusine/server/key"
        tmp_key_file = "/var/lib/debusine/server/key.tmp"

        shutil.move(key_file, tmp_key_file)
        self.addCleanup(DebusineServer.restart)
        self.addCleanup(shutil.move, tmp_key_file, key_file)

        # Cannot start because key file cannot be read
        self.assertEqual(DebusineServer.restart().returncode, 1)

    def test_worker_is_not_connected(self) -> None:
        """The worker is not connected: the token is not enabled yet."""
        self._worker.set_up(use_activation_token=False)

        self.assertTrue(
            DebusineServer.verify_worker(
                self._worker.read_token(),
                connected=False,
                enabled=False,
            )
        )

    def test_worker_connects(self) -> None:
        """The admin enables the worker and it gets connected."""
        self.connect_worker()

        self.assertTrue(
            DebusineServer.verify_worker(
                Worker.read_token(), connected=True, enabled=True
            )
        )

    def test_create_work_requests_invalid(self) -> None:
        """Create an invalid work request."""
        # Client submits an invalid work-request (status will be 'error')
        # It is not valid because task_data is not valid

        self.connect_worker()

        create_work_request_output = Client.execute_command(
            "work-request", "create", "--yaml", self.TASK_NAME, stdin="foo: bar"
        )

        # Work Request is registered
        self.assertEqual(create_work_request_output["result"], "failure")

        # Specific "title" and "detail" are tested in the unit tests
        self.assertIn("title", create_work_request_output["error"])
        self.assertIn("detail", create_work_request_output["error"])

    def test_create_work_request_valid(self) -> None:
        """Create a valid work request."""
        self.connect_worker()

        task_data = textwrap.dedent(
            '''\
                result:
                    success
                '''
        )
        self.assert_create_work_request(
            task_data=task_data,
            expected_status="completed",
            expected_result="success",
        )

    def test_debusine_admin_print_permission_error_message(self) -> None:
        """debusine-admin is invoked with wrong user: print error message."""
        # This test is to catch invalid permission
        # (in the logs files for example). Since the test for the secret_key
        # happens before we change the secret_key permissions so it fails
        # later.
        secret_key = "/var/lib/debusine/server/key"
        self.addCleanup(shutil.chown, secret_key, Path(secret_key).owner())
        temp_user = "postgres"
        shutil.chown(secret_key, temp_user)

        result = DebusineServer.execute_command("list_tokens", user=temp_user)

        self.assertEqual(
            "Permission error: Unable to configure handler 'debug.log'. "
            "Check that the user running debusine-admin has access to the"
            " file \"/var/log/debusine/server/debug.log\".\n",
            result.stderr,
        )
        self.assertEqual(result.returncode, 3)

    def assert_create_work_request(
        self, *, task_data: str, expected_status: str, expected_result: str
    ) -> None:
        """
        Submit via debusine client a work request to the server.

        Create work request, assert creation and wait for the work request
        to be completed. Assert expected_status and expected_result.
        """
        original_start_latency_count = DebusineServer.get_metric(
            metric_name="debusine_work_request_start_latency",
            sample_name="debusine_work_request_start_latency_count",
            task_type="Worker",
            task_name=self.TASK_NAME,
            priority="normal",
        )

        create_work_request_output = Client.execute_command(
            "work-request", "create", "--yaml", self.TASK_NAME, stdin=task_data
        )

        work_request_id = create_work_request_output["id"]

        # The client will check the status of the created request
        show_work_request_output = Client.execute_command(
            "work-request", "show", work_request_id
        )

        # Task name is as submitted
        self.assertEqual(show_work_request_output['task_name'], self.TASK_NAME)

        # Task data is as submitted
        actual_task_data = show_work_request_output['task_data'].copy()
        self.assertIsInstance(actual_task_data.pop("task_configuration"), int)
        self.assertEqual(actual_task_data, yaml.safe_load(task_data))

        # The worker should get the new work request and start executing it
        Client.wait_for_work_request_completed(work_request_id, expected_result)

        # Assert expected_result and expected_status
        show_work_request_output = Client.execute_command(
            "work-request", "show", work_request_id
        )

        self.assertEqual(show_work_request_output["status"], expected_status)
        self.assertEqual(show_work_request_output["result"], expected_result)

        # Metrics are recorded
        new_start_latency_count = DebusineServer.get_metric(
            metric_name="debusine_work_request_start_latency",
            sample_name="debusine_work_request_start_latency_count",
            task_type="Worker",
            task_name=self.TASK_NAME,
            priority="normal",
        )
        self.assertEqual(
            new_start_latency_count, (original_start_latency_count or 0.0) + 1
        )

    def test_web_login(self) -> None:
        """
        Ensure the web interface is accessible after login.

        Catches unit-tests-escaping issues like
        https://salsa.debian.org/freexian-team/debusine/-/issues/563
        """
        s = requests.session()
        r = s.get(f"{Configuration.get_base_url()}/-/login")
        html = lxml.html.fromstring(r.content)
        csrftoken = html.xpath(
            """//input[@name="csrfmiddlewaretoken"]/@value"""
        )[0]
        with open(os.environ["AUTOPKGTEST_TMP"] + "/test-password.txt") as f:
            password = f.read()[:-1]
        r = s.post(
            f"{Configuration.get_base_url()}/-/login/",
            {
                "username": "test-user",
                "password": password,
                "csrfmiddlewaretoken": csrftoken,
            },
            # Referer required in https
            headers={"Referer": f"{Configuration.get_base_url()}/-/login"},
        )
        html = lxml.html.fromstring(r.content)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(
            html.xpath("""//ul[@id='navbar-right']/li/a/text()""")[0].strip(),
            "test-user",
        )
